# -*- coding: utf-8 -*-
import os
import pytest


def test_resolver_check_candidate_ok_001(settings, parser, resolver):
    """resolver.ImportPathsResolver: Check candidates correct case 1"""
    candidates = resolver.candidate_paths("vendor")
    
    results = resolver.check_candidate_exists(settings.sample_path, candidates)
    
    assert results == os.path.join(settings.sample_path, "_vendor.scss")


def test_resolver_check_candidate_ok_002(settings, parser, resolver):
    """resolver.ImportPathsResolver: Check candidates correct case 2"""
    candidates = resolver.candidate_paths("components/_filename_test_2")
    
    results = resolver.check_candidate_exists(settings.sample_path, candidates)
    
    assert results == os.path.join(settings.sample_path, "components/_filename_test_2.scss")


def test_resolver_check_candidate_ok_003(settings, parser, resolver):
    """resolver.ImportPathsResolver: Check candidates correct case 3"""
    candidates = resolver.candidate_paths("components/filename_test_6.plop.scss")
    
    results = resolver.check_candidate_exists(settings.sample_path, candidates)
    
    assert results == os.path.join(settings.sample_path, "components/_filename_test_6.plop.scss")


def test_resolver_check_candidate_ok_004(settings, parser, resolver):
    """resolver.ImportPathsResolver: Check candidates correct case 4"""
    basepath = os.path.join(settings.sample_path, "components")
    candidates = resolver.candidate_paths("webfont")
    
    results = resolver.check_candidate_exists(basepath, candidates)
    
    assert results == os.path.join(basepath, "_webfont.scss")


def test_resolver_check_candidate_ok_005(settings, parser, resolver):
    """resolver.ImportPathsResolver: Check candidates correct case 5"""
    basepath = os.path.join(settings.sample_path, "components")
    candidates = resolver.candidate_paths("../components/webfont_icons")
    
    results = resolver.check_candidate_exists(basepath, candidates)
    
    assert results == os.path.join(basepath, "../components/_webfont_icons.scss")


def test_resolver_check_candidate_wrong_001(settings, parser, resolver):
    """resolver.ImportPathsResolver: Check candidates wrong case 1"""
    candidates = resolver.candidate_paths("dont_exists")
    
    results = resolver.check_candidate_exists(settings.sample_path, candidates)
    
    assert results == False


def test_resolver_check_candidate_wrong_002(settings, parser, resolver):
    """resolver.ImportPathsResolver: Check candidates wrong case 2"""
    candidates = resolver.candidate_paths("css_filetest.sass")
    
    results = resolver.check_candidate_exists(settings.sample_path, candidates)
    
    assert results == False



"""
Some fixture methods 
"""
import os
import pytest

import boussole
from boussole.parser import ScssImportsParser
from boussole.resolver import ImportPathsResolver
from boussole.inspector import ScssInspector

class FixturesSettingsTestMixin(object):
    """Mixin containing some basic settings"""
    def __init__(self):
        # Base fixture datas directory
        self.fixtures_dir = 'data_fixtures'
        self.fixtures_path = os.path.normpath(
            os.path.join(
                os.path.abspath(os.path.dirname(boussole.__file__)),
                '..',
                'tests',
                self.fixtures_dir
            )
        )
        
        # Sample project
        self.sample_dir = "sample_project"
        self.sample_path = os.path.join(self.fixtures_path, self.sample_dir)
        
        # Some sample libraries
        self.lib1_dir = 'library_1'
        self.lib2_dir = 'library_2'
        self.lib1_path = os.path.join(self.fixtures_path, self.lib1_dir)
        self.lib2_path = os.path.join(self.fixtures_path, self.lib2_dir)
        self.libraries_fixture_paths = [
            self.lib1_path,
            self.lib2_path,
        ]


@pytest.fixture(scope="module")
def settings():
    """Initialize and return settings (mostly paths) for fixtures (scope at module level)"""
    return FixturesSettingsTestMixin()

@pytest.fixture(scope="module")
def parser():
    """Initialize and return SCSS parser (scope at module level)"""
    return ScssImportsParser()

@pytest.fixture(scope="module")
def resolver():
    """Initialize and return Path resolver (scope at module level)"""
    return ImportPathsResolver()

@pytest.fixture(scope="function")
def inspector():
    """Initialize and return SCSS inspector (scope at function level)"""
    return ScssInspector()



# -*- coding: utf-8 -*-
"""
Parser
======

.. _SASS Reference:
    http://sass-lang.com/documentation/file.SASS_REFERENCE.html#import
"""
import re

from boussole.exceptions import InvalidImportRule


class ScssImportsParser(object):
    """
    SCSS parser to find import rules.

    Builded from `SASS Reference`_ about "@import" rule.

    This does not support the old SASS syntax (also known as "indented
    syntax").

    It's a mixin, meaning without own ``__init__`` method so it's should be
    safe enough to inherit it from another class.

    Attributes:
        REGEX_IMPORT_RULE: Compiled regex used to find import rules.
        REGEX_COMMENTS: Compiled regex used to find and remove comments.
    """
    REGEX_IMPORT_RULE = re.compile(ur'@import\s*(url)?\s*\(?([^;]+?)\)?;',
                                   re.IGNORECASE)
    REGEX_COMMENTS = re.compile(r'(/\*.*?\*/)|(//.*?(\n|$))',
                                re.IGNORECASE | re.DOTALL)

    def strip_quotes(self, content):
        """
        Unquote given rule.

        Args:
            content (str): An import rule.

        Raises:
            InvalidImportRule: Raise exception if the rule is badly quoted
            (not started or not ended quotes).

        Returns:
            string: The given rule unquoted.
        """
        error_msg = "Following rule is badly quoted: {}"
        if (content.startswith('"') and content.endswith('"')) or \
           (content.startswith("'") and content.endswith("'")):
            return content[1:-1]
        # Quote starting but not ended
        elif (content.startswith('"') and not content.endswith('"')) or \
             (content.startswith("'") and not content.endswith("'")):
            raise InvalidImportRule(error_msg.format(content))
        # Quote ending but not started
        elif (not content.startswith('"') and content.endswith('"')) or \
             (not content.startswith("'") and content.endswith("'")):
            raise InvalidImportRule(error_msg.format(content))

        return content

    def remove_comments(self, content):
        """
        Remove all comment kind (inline and multiline) from given content.

        Args:
            content (str): A SCSS source.

        Returns:
            string: Given SCSS source with all comments removed.
        """
        return self.REGEX_COMMENTS.sub("", content)

    def filter_rules(self, path):
        """
        Lambda to filter items that:
        * Starts with http:// or https:// (this for external load only)
        * Ends with ".css" (they are not intended to be compiled)
        """
        return not(path.startswith('http://') or
                   path.startswith('https://') or path.endswith('.css'))

    def flatten_rules(self, declarations):
        """
        Flatten returned import rules from regex.

        Because import rules can contains multiple items in the same rule
        (called multiline import rule), the regex ``REGEX_IMPORT_RULE``
        return a list of unquoted items for each rule.

        Args:
            declarations (list): A SCSS source.

        Returns:
            list: Given SCSS source with all comments removed.
        """
        rules = []

        for protocole, paths in declarations:
            # If there is a protocole (like 'url), drop it
            if protocole:
                continue
            # Unquote and possibly split multiple rule in the same declaration
            rules.extend([self.strip_quotes(v.strip())
                          for v in paths.split(',')])

        return filter(self.filter_rules, rules)

    def parse(self, content):
        """
        Parse a stylesheet document with a regex (``REGEX_IMPORT_RULE``)
        to extract all import rules and return them.

        Args:
            content (str): A SCSS source.

        Returns:
            list: Finded paths in import rules.
        """
        # Remove all comments before searching for import rules, to not catch
        # commented breaked import rules
        declarations = self.REGEX_IMPORT_RULE.findall(
            self.remove_comments(content)
        )
        return self.flatten_rules(declarations)



# -*- coding: utf-8 -*-
"""
Resolver
========

.. _SASS @import Reference:
    http://sass-lang.com/documentation/file.SASS_REFERENCE.html#import

.. _SASS partials Reference:
    http://sass-lang.com/documentation/file.SASS_REFERENCE.html#partials
"""
import os

from boussole.exceptions import UnresolvablePath


class ImportPathsResolver(object):
    """
    Import paths resolver.

    Resolve given paths from SCSS source to absolute paths.

    Example:
        These rules will be resolved: ::

            // Real filename: "components/_header.scss"
            @import "components/header";
            @import "components/_header";
            @import "components/header.scss";
            @import "components/_header.scss";

    It's a mixin, meaning without own ``__init__`` method so it's should be
    safe enough to inherit it from another class.

    Attributes:
        CANDIDATE_EXTENSIONS (list): List of extensions available to build
            candidate paths. Beware, order does matter, the first extension
            will be the top candidate.
        STRICT_PATH_VALIDATION (bool): A switch to enabled (``True``) or
            disable (``False``) exception raising when a path can not be
            resolved.
    """
    CANDIDATE_EXTENSIONS = ['scss', 'sass', 'css', ]
    STRICT_PATH_VALIDATION = True

    def candidate_paths(self, filepath):
        """
        Return candidates path for given path

        * If Filename does not starts with ``_``, will build a candidate for
          both with and without ``_`` prefix;
        * Will build For each available extensions if filename does not have
          an explicit extension;
        * Leading path directory is preserved;

        Args:
            filepath (str): Relative path as finded in an import rule from a
                SCSS source.

        Returns:
            list: Builded candidate paths (as relative paths).
        """
        filelead, filetail = os.path.split(filepath)
        name, extension = os.path.splitext(filetail)
        # Removed leading dot from extension
        if extension:
            extension = extension[1:]

        filenames = [name]
        # If underscore prefix is present, dont need to double underscore
        if not name.startswith('_'):
            filenames.append("_{}".format(name))

        # If explicit extension, dont need to add more candidate extensions
        if extension and extension in self.CANDIDATE_EXTENSIONS:
            filenames = [".".join([k, extension]) for k in filenames]
        # Else if no extension or not candidate, add candidate extensions
        else:
            # Restore uncandidate extensions if any
            if extension:
                filenames = [".".join([k, extension]) for k in filenames]
            new = []
            for ext in self.CANDIDATE_EXTENSIONS:
                new.extend([".".join([k, ext]) for k in filenames])
            filenames = new

        # Return candidates with restored leading path if any
        return [os.path.join(filelead, v) for v in filenames]

    def check_candidate_exists(self, basepath, candidates):
        """
        Check that at least one candidate exist into a directory.

        Args:
            basepath (str): Directory path where to search for candidate.
            candidates (list): List of candidate file paths.

        Returns:
            ``False`` if no candidate exists else return the elected path.
        """
        for item in candidates:
            abspath = os.path.join(basepath, item)
            if os.path.exists(abspath):
                return abspath

        return False

    def resolve(self, sourcepath, paths, library_paths=None):
        """
        Resolve given paths from given base paths

        Return resolved path list.

        Note:
            Resolving strategy is made like libsass do, meaning paths in
            import rules is resolved from the source where the import rules
            have been finded.

        Args:
            sourcepath (str): Source file path, its directory is used to
                resolve given paths. The path must be an absolute path to
                avoid errors on resolving.
            paths (list): Relative paths (from ``sourcepath``) to resolve.
            library_paths (list): List of directory paths for libraries to
                resolve paths if resolving fails on the base source path.
                Default to None.

        Raises:
            UnresolvablePath: If a path does not exist and
                ``STRICT_PATH_VALIDATION`` attribute is ``True``.

        Returns:
            list: List of resolved path.
        """
        # Split basedir/filename from sourcepath, so the first resolving
        # basepath is the sourcepath directory, then the optionnal
        # given libraries
        basedir, filename = os.path.split(sourcepath)
        basepaths = [basedir]
        resolved_paths = []

        # Add given library paths to the basepaths for resolving
        # Accept a string if not allready in basepaths
        if library_paths and isinstance(library_paths, basestring) and \
           library_paths not in basepaths:
            basepaths.append(library_paths)
        # Add path item from list if not allready in basepaths
        elif library_paths:
            for k in list(library_paths):
                if k not in basepaths:
                    basepaths.append(k)

        for import_rule in paths:
            candidates = self.candidate_paths(import_rule)

            existing = False
            # The first resolved candidate from basepaths wins
            for i, basepath in enumerate(basepaths):
                existing = self.check_candidate_exists(basepath, candidates)
                if existing:
                    # Normalized resolved path
                    resolved_paths.append(os.path.normpath(existing))
                    break

            if not existing and self.STRICT_PATH_VALIDATION:
                raise UnresolvablePath(
                    "Imported path '{}' does not exist in '{}'".format(
                        import_rule, basedir
                    )
                )

        return resolved_paths



# -*- coding: utf-8 -*-
# flake8: noqa
"""
Inspector development stuff

DO NOT COMMIT
"""
import os
import json
import boussole

from boussole.exceptions import CircularImport
from boussole.inspector import ScssInspector

class CatchedRuntimeError(CircularImport):
    pass

def shortpath(path, base_dir):
    if path.startswith(base_dir):
        return path[len(base_dir):]
    return path

def shorten_paths_map(paths_map, base_dir):
    new_map = {}
    for k,v in paths_map.items():
        new_map[shortpath(k, base_dir)] = [shortpath(item, base_dir) for item in v]
    return new_map

def shorten_paths_list(paths_list, base_dir):
    return [shortpath(k, base_dir) for k in paths_list]
    
class ComplexEncoder(json.JSONEncoder):
    """
    Allow to encode properly some objet types
    """
    def default(self, obj):
        if isinstance(obj, set):
            return list(obj)
        return json.JSONEncoder.default(self, obj)

# For some development debug
if __name__ == "__main__":
    
    boussole_dir = os.path.dirname(boussole.__file__)
    fixtures_dir = os.path.normpath(os.path.join(os.path.abspath(boussole_dir), '..', 'tests', 'data_fixtures'))
    sample_path = os.path.join(fixtures_dir, 'sample_project')
    library1_path = os.path.join(fixtures_dir, 'library_1')
    library2_path = os.path.join(fixtures_dir, 'library_2')
    library_paths = [library1_path, library2_path]
    
    inspector = ScssInspector()
    
    sources = [
        os.path.join(sample_path, 'main_basic.scss'),
        os.path.join(sample_path, 'main_syntax.scss'),
        os.path.join(sample_path, 'main_commented.scss'),
        os.path.join(sample_path, 'main_depth_import-1.scss'),
        os.path.join(sample_path, 'main_depth_import-2.scss'),
        os.path.join(sample_path, 'main_depth_import-3.scss'),
        os.path.join(sample_path, 'main_with_subimports.scss'),
        os.path.join(sample_path, 'main_using_libs.scss'),
        os.path.join(sample_path, 'main_circular_0.scss'),
        os.path.join(sample_path, 'main_circular_1.scss'),
        os.path.join(sample_path, 'main_circular_2.scss'),
        os.path.join(sample_path, 'main_circular_3.scss'),
        os.path.join(sample_path, 'main_circular_4.scss'),
        os.path.join(sample_path, 'main_circular_bridge.scss'),
        os.path.join(sample_path, 'main_circular_5.scss'),
    ]
    
    basic_sourcepath = os.path.join(sample_path, 'main_basic.scss')
    depth1_sourcepath = os.path.join(sample_path, 'main_depth_import-1.scss')
    depth2_sourcepath = os.path.join(sample_path, 'main_depth_import-2.scss')
    depth3_sourcepath = os.path.join(sample_path, 'main_depth_import-3.scss')
    subimports_sourcepath = os.path.join(sample_path, 'main_with_subimports.scss')
    usinglibs_sourcepath = os.path.join(sample_path, 'main_using_libs.scss')
    circular0_sourcepath = os.path.join(sample_path, 'main_circular_0.scss')
    circular1_sourcepath = os.path.join(sample_path, 'main_circular_1.scss')
    circular2_sourcepath = os.path.join(sample_path, 'main_circular_2.scss')
    circular3_sourcepath = os.path.join(sample_path, 'main_circular_3.scss')
    circular4_sourcepath = os.path.join(sample_path, 'main_circular_4.scss')
    circularbridge_sourcepath = os.path.join(sample_path, 'main_circular_bridge.scss')
    circular5_sourcepath = os.path.join(sample_path, 'main_circular_5.scss')
    vendor_sourcepath = os.path.join(sample_path, '_vendor.scss')
    webfont_sourcepath = os.path.join(sample_path, 'components/_webfont.scss')
    icons_sourcepath = os.path.join(sample_path, 'components/_webfont_icons.scss')
    
    results = inspector.inspect(*sources, library_paths=library_paths)
    
    # Display maps from inspection
    print "-"*80
    print "Sources that import other sources (_CHILDREN_MAP)"
    print "-"*80
    
    print json.dumps(shorten_paths_map(inspector._CHILDREN_MAP, sample_path+'/'), indent=4)
    print
    
    
    print
    print "-"*80
    print "Sources that are imported by other sources (_PARENTS_MAP)"
    print "-"*80
    
    print json.dumps(shorten_paths_map(inspector._PARENTS_MAP, sample_path+'/'), indent=4, cls=ComplexEncoder)
    print
    
    print "~"*200
    
    def display_children_process(sourcepath):
        print
        print "-"*80
        print "(inspector.children) They are imported by:", shortpath(sourcepath, sample_path+'/')
        print "-"*80
        
        try:
            deps = list(inspector.children(sourcepath))
        except CatchedRuntimeError:
            print "[!] CatchedRuntimeError error has been raised [!]"
        except CircularImport:
            print "[!] CircularImport error has been raised [!]"
        else:
            print json.dumps(shorten_paths_list(deps, sample_path+'/'), indent=4)
            print
    
    def display_parents_process(sourcepath):
        print
        print "-"*80
        print "(inspector.parents) They are imported by:", shortpath(sourcepath, sample_path+'/')
        print "-"*80
        
        try:
            deps = list(inspector.parents(sourcepath))
        except CatchedRuntimeError:
            print "[!] CatchedRuntimeError error has been raised [!]"
        except CircularImport:
            print "[!] CircularImport error has been raised [!]"
        else:
            print json.dumps(shorten_paths_list(deps, sample_path+'/'), indent=4)
            print
    
    
    ## Basic sample
    #display_children_process(basic_sourcepath)
    
    #print "~"*200
    
    ## Depth sample
    #display_children_process(depth3_sourcepath)
    
    #print "~"*200
    
    ## 
    #display_children_process(usinglibs_sourcepath)
    
    #print "~"*200
    
    ## Some test for case 1>2
    #display_children_process(circular1_sourcepath)
    
    #print "~"*200
    
    ## Some test for case 3>bridge>4
    #display_children_process(circular3_sourcepath)
    
    #print "~"*200
    
    ## Final case for sub circular import
    #display_children_process(circular5_sourcepath)
    
    #print "~"*200
    
    ## Self circular import
    #display_children_process(circular0_sourcepath)
    
    print "~"*200
    
    # ...
    display_parents_process(vendor_sourcepath)
    display_parents_process(icons_sourcepath)



# -*- coding: utf-8 -*-
import os
import pytest

def test_inspector_parents_basic(settings, inspector):
    """inspector.ScssInspector: Looking for parents of basic sample"""
    sources = [
        os.path.join(settings.sample_path, 'main_basic.scss'),
        os.path.join(settings.sample_path, 'main_depth_import-3.scss'),
        os.path.join(settings.sample_path, 'main_with_subimports.scss'),
        os.path.join(settings.sample_path, 'main_using_libs.scss'),
    ]
    sourcepath = os.path.join(settings.sample_path, 'main_basic.scss')
    
    inspector.inspect(*sources, library_paths=settings.libraries_fixture_paths)

    parents = list(inspector.parents(sourcepath))
    assert parents == [
        os.path.join(settings.sample_path, 'main_depth_import-1.scss'),
        os.path.join(settings.sample_path, 'main_depth_import-2.scss'),
        os.path.join(settings.sample_path, 'main_depth_import-3.scss'),
        os.path.join(settings.sample_path, 'main_with_subimports.scss'),
        os.path.join(settings.sample_path, 'main_using_libs.scss'),
    ]

def test_inspector_parents_vendor(settings, inspector):
    """inspector.ScssInspector: Looking for parents of vendor component"""
    sources = [
        os.path.join(settings.sample_path, 'main_syntax.scss'),
        os.path.join(settings.sample_path, 'main_commented.scss'),
        os.path.join(settings.sample_path, 'main_basic.scss'),
        os.path.join(settings.sample_path, 'main_depth_import-1.scss'),
        os.path.join(settings.sample_path, 'main_depth_import-2.scss'),
        os.path.join(settings.sample_path, 'main_depth_import-3.scss'),
        os.path.join(settings.sample_path, 'main_with_subimports.scss'),
        os.path.join(settings.sample_path, 'main_using_libs.scss'),
        os.path.join(settings.sample_path, 'main_circular_0.scss'),
        os.path.join(settings.sample_path, 'main_circular_1.scss'),
        os.path.join(settings.sample_path, 'main_circular_2.scss'),
        os.path.join(settings.sample_path, 'main_circular_3.scss'),
        os.path.join(settings.sample_path, 'main_circular_4.scss'),
        os.path.join(settings.sample_path, 'main_circular_bridge.scss'),
        os.path.join(settings.sample_path, 'main_circular_5.scss'),
    ]
    sourcepath = os.path.join(settings.sample_path, '_vendor.scss')
    
    inspector.inspect(*sources, library_paths=settings.libraries_fixture_paths)

    parents = list(inspector.parents(sourcepath))
    assert parents == [
        os.path.join(settings.sample_path, 'main_depth_import-1.scss'),
        os.path.join(settings.sample_path, 'main_depth_import-2.scss'),
        os.path.join(settings.sample_path, 'main_circular_4.scss'),
        os.path.join(settings.sample_path, 'main_circular_5.scss'),
        os.path.join(settings.sample_path, 'main_circular_bridge.scss'),
        os.path.join(settings.sample_path, 'main_commented.scss'),
        os.path.join(settings.sample_path, 'main_depth_import-3.scss'),
        os.path.join(settings.sample_path, 'main_with_subimports.scss'),
        os.path.join(settings.sample_path, 'main_basic.scss'),
        os.path.join(settings.sample_path, 'main_syntax.scss'),
        os.path.join(settings.sample_path, 'main_circular_3.scss'),
        os.path.join(settings.sample_path, 'main_using_libs.scss'),
    ]



# -*- coding: utf-8 -*-
import os
import pytest

from boussole.exceptions import UnresolvablePath

def test_resolver_resolve_basic(settings, parser, resolver):
    """resolver.ImportPathsResolver: Resolve paths from basic sample"""
    sourcepath = os.path.join(settings.sample_path, 'main_basic.scss')
    with open(sourcepath) as fp:
        finded_paths = parser.parse(fp.read())
        
    resolved_paths = resolver.resolve(sourcepath, finded_paths)
    
    assert resolved_paths == [
        os.path.join(settings.sample_path, '_vendor.scss'),
        os.path.join(settings.sample_path, '_empty.scss'),
    ]


def test_resolver_resolve_library(settings, parser, resolver):
    """resolver.ImportPathsResolver: Resolve paths from main_using_libs.scss 
    that use included libraries"""
    sourcepath = os.path.join(settings.sample_path, 'main_using_libs.scss')
    with open(sourcepath) as fp:
        finded_paths = parser.parse(fp.read())
    
    resolved_paths = resolver.resolve(
        sourcepath, 
        finded_paths, 
        library_paths=settings.libraries_fixture_paths
    )
    
    assert resolved_paths == [
        os.path.join(settings.lib2_path, 'addons/_some_addon.scss'),
        os.path.join(settings.sample_path, 'main_basic.scss'),
        os.path.join(settings.sample_path, 'components/_webfont.scss'),
        os.path.join(settings.lib1_path, 'library_1_fullstack.scss'),
    ]


def test_resolver_resolve_commented(settings, parser, resolver):
    """resolver.ImportPathsResolver: Resolve paths from sample with comments"""
    sourcepath = os.path.join(settings.sample_path, 'main_commented.scss')
    with open(sourcepath) as fp:
        finded_paths = parser.parse(fp.read())
    
    resolved_paths = resolver.resolve(sourcepath, finded_paths)
    
    assert resolved_paths == [
        os.path.join(settings.sample_path, '_vendor.scss'),
        os.path.join(settings.sample_path, 'components/_filename_test_1.scss'),
        os.path.join(settings.sample_path, '_empty.scss'),
    ]


def test_resolver_resolve_error(settings, parser, resolver):
    """resolver.ImportPathsResolver: Exception on wrong import path"""
    sourcepath = os.path.join(settings.sample_path, 'main_error.scss')
    with open(sourcepath) as fp:
        finded_paths = parser.parse(fp.read())
        
    with pytest.raises(UnresolvablePath):
        resolver.resolve(
            sourcepath, 
            finded_paths, 
            library_paths=settings.libraries_fixture_paths
        )
